/**
   Teensy32AnalogLogger is a modification of SdFat AnalogBinLogger. https://github.com/greiman/SdFat
   modications by Matthew Ford
   (c)2019 Forward Computing and Control Pty. Ltd. NSW Australia
   All rights reserverd.
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.

   This sketch uses two AnalogToDigital converter modules, ADC_0 and ADC_1, to
   sample two signals at the same time. This allows for higher speed sampling.
   However on the Teensy 3.2 (and 3.5 and 3.6) not all Analog inputs can be
   sampled by ADC_0 and/or ADC_1
   The Teensy3_1_AnalogCard.png image in this sketch directory shows which
   pins can be read by which ADC_x for Teensy 3.1 and Teensy 3.2
   Here only single end reads are considered. Teensy can also do differential ADC on some pins

   For Single Ended Reads i.e. +Volts referenced to GND
   ADC_0 can read A0 to A9, A10, A11, A12, A14
   ADC_1 can read A2, A3, A10, A13, A15 to A20
   If you select a pin that the ADC cannot read it will return 0 (always)

   Communication is via Serial1 to bluetooth module at 115200

*/

/**
   from https://github.com/greiman/SdFat AnalogBinLogger
   This program logs data from the Arduino ADC to a binary file.

   Samples are logged at regular intervals. Each Sample consists of the ADC
   values for the analog pins defined in the PIN_LIST array.  The pins numbers
   may be in any order.

   Edit the configuration constants below to set the sample pins, sample rate,
   and other configuration values.

   If your SD card has a long write latency, it may be necessary to use
   slower sample rates.  Using a Mega Arduino helps overcome latency
   problems since 13 512 byte buffers will be used.

   Each 512 byte data block in the file has a four byte header followed by up
   to 508 bytes of data. (508 values in 8-bit mode or 254 values in 10-bit to 16-bit mode)
   Each block contains an integral number of samples with unused space at the
   end of the block.

   Data is written to the file using a SD multiple block write command.
*/

/**
   Outline of code
   Setup initializes buffers via adcInit populates metadataBuffer
   In loop()

*/

#ifdef __MK20DX256__
// teensy 3.2 only
#include <SPI.h>
#include "ADC.h"
#include "SdFat.h"
#include "FreeStack.h"
#include "Teensy32AnalogLogger.h"
#include <avr/sleep.h>

// IntervalTimer is part of teensy 3.2 hardware support
#include <IntervalTimer.h>

// from https://www.forward.com.au/pfod/ArduinoProgramming/TimingDelaysInArduino.html
// library https://www.forward.com.au/pfod/ArduinoProgramming/millisDelay.zip
#include <millisDelay.h>

// from https://www.forward.com.au/pfod/pfodParserLibraries/index.html
// library https://www.forward.com.au/pfod/pfodParserLibraries/pfodParser.zip
#include <pfodBufferedStream.h>
#include <pfodNonBlockingInput.h>

const uint32_t versionByteOrder =  0x0000ffff; // versionNo + ffff versionNo is the metadata version
// SUB_VERSION is the Teensy32AnalogLogger code version
#define SUB_VERSION ".02"

// use Serial for USB control
// OR  Serial1 for Commuication board connected D0/D1
#define COM_SERIAL Serial1
const uint32_t baudRate = 115200;
Stream* serialIO; // where the buffered IO is sent to and read from

// if this next line is un-commented then get some startup debug output to Serial at 115200
//#define DEBUG

//------------------------------------------------------------------------------
// Analog pin number list for a sample.   For Single Ended Reads i.e. +Volts referenced to GND
//   ADC_0 can read A0 to A9, A10, A11, A12, A14
//   ADC_1 can read A2, A3, A10, A13, A15 to A20
//   If you select a pin that the ADC cannot read it will return 0 (always)
const uint8_t PIN_LIST[] = {A2, A3}; //i.e. pin for ADC_0 to read, pin ADC_1 to read

const int SAMPLE_INTERVAL = 33;//1000; // us == 1mS == 1000 /sec  at 25uS -> 40K 2xsamples/sec miss timer every now and again
const uint8_t BUFFER_BLOCK_COUNT = 70;  // must be <=124 see QUEUE_DIM below

const uint8_t ADC_AVGS = 4; //0, 4, 8, 16 or 32 This averaging happens in hardware before the reading is returned
const uint8_t ADC_BITS = 10; // 8, 10, 12 or 16.  8 and 10 bit guarenteed to be monotonic with no missing bits
// Note if using 8 bit, then should recode to only store 8 i.e. use DATA_DIM8 / block8_t

unsigned long RUN_TIME_MS = 30ul * 24 * 60 * 60 * 1000; //i.e. 30days in mS OR 30000 for 30sec testing
// may be reduced if not enough files to save data

const unsigned long PRINT_INTERVAL = 10000; // mS interval to print progress message while logging

// run stops after this time OR when all files are full
// max number of files to write max is 99 ~12Gig
// unused files are removed after logging stops
// set to 3 for testing  normally set to 99
#define FILE_COUNT 99


// PROTECT_DATA prevents overwriting previous run's log files.  Can set to false for testing
const bool PROTECT_DATA = true;  // this prevents overrite of log files after first run

// Maximum file size in blocks.
// The program creates a contiguous files with FILE_BLOCK_COUNT 512 byte blocks.
//  The file will be truncated if logging is stopped early and unused files removed
const uint32_t FILE_BLOCK_COUNT = 256000;  // file size is FILE_BLOCK_COUNT/2 Kb

// unsigned long == uint32_t on teensy 3.2

void closeBinFileForBlockWrite();
void timer0_callback(void);
void adc1_isr();
void saveSample();
void fatalBlink();
void initializeQueues(); // sets empty queue and clears full queue
void adcInit(metadata_t* meta);
void adcStart();
void adcStop();
void mStoHrMinSec(unsigned long mS);
int setNextBinName(int count);
void closeBinFileForBlockWrite();
void openBinFileForBlockWrite();
bool createEmptyFile();
bool metadataFileExists();
bool initializeFiles();
void printMenu();
void checkErrorBlink();

void outputMetaDataToSerial(metadata_t* meta);
void updateMetaData(metadata_t* metaBuffer, bool updateDuration);
bool updateMetaDataFile(metadata_t* metaBuffer);
void readMetaDataFile(metadata_t* metaBuffer);
bool writeMetadataFile(metadata_t* metaBuffer); // creates new metaFile if it does not exist

void storeLastBuffersAndTruncate();
void processingOnADCstop();

void setup(void);
void processRunCmd();
void handleInput(char c);
char checkInputForCmd();
void loop(void);

// for readInputLine
const size_t lineBufLen = 81; // 80+null
char lineBuf[lineBufLen] = ""; // start empty

SdFat sd;  // standard SPI to SD

const size_t STREAM_BUFFER_SIZE = 5120;
uint8_t streamBuffer[STREAM_BUFFER_SIZE];
// these two classes are from pfodParser library  https://www.forward.com.au/pfod/pfodParserLibraries/index.html
pfodBufferedStream bufferedStream(baudRate, streamBuffer, STREAM_BUFFER_SIZE, false); // non blocking will drop any excess bytes
pfodNonBlockingInput nonBlocking;
uint16_t version; // set in setup() from versionByteOrder
//==============================================================================
// Error messages stored in flash.
#define error(msg) {sd.errorPrint(F(msg));fatalBlink();}

IntervalTimer timer0;
ADC *adc = new ADC(); // adc object
millisDelay printTimer;
millisDelay stopTimer; // set run time with stop timer else stop when files all full

#if (FILE_COUNT > 99)
#error File Count can not exceed 99
#endif

const uint8_t PIN_COUNT = sizeof(PIN_LIST) / sizeof(PIN_LIST[0]); // number of pins to log ALWAYS 2 here
const size_t SAMPLES_PER_BLOCK = DATA_DIM16 / PIN_COUNT;
typedef block16_t block_t;
//------------------------------------------------------------------------------

uint32_t maxLatency = 0; // max time in uS to save a single block to SD
// max number of blocks to erase per erase call
uint32_t const ERASE_SIZE = 262144L;
unsigned long mSstart; // start time
int fileCounter = 0; // index of current file
bool haveAlreadyLoggedData = false; // run data logging only once

// log file base name.  Must be six characters or less.
#define FILE_BASE_NAME "log"
// usually use FILE_BASE_EXTRA  "00.bin"
#define FILE_BASE_EXTRA  "00.bin"
#define FILE_BASE_META_EXTRA  ".met"

// Size of file base name.  Must not be larger than six.
const uint8_t BASE_NAME_SIZE = sizeof(FILE_BASE_NAME) - 1;

//------------------------------------------------------------------------------
// Pin definitions.
//
// Digital pin to indicate an error, set to -1 if not used.
// The led blinks for fatal errors.
// The led goes on solid for SD write missedSamples errors (missed samples) and logging continues.
const int8_t ERROR_LED_PIN = 3;
const int8_t ERROR_LED_GND_PIN = 2; // set this low to supply GND for error led, set to -1 if not used for a ground
const int8_t MONITOR_PIN = -1; // 4; // set to -1 for not used
bool fatalError = false;
bool loggingRunning = false;

uint32_t requestedRunTime = RUN_TIME_MS; // may be less if exceeds FILE_COUNT files

// SD chip select pin.
const uint8_t SD_CS_PIN = SS;
//------------------------------------------------------------------------------
// Buffer definitions.
//
// The logger will use SdFat's buffer plus BUFFER_BLOCK_COUNT additional
// buffers.  QUEUE_DIM must be a power of two larger than
//(BUFFER_BLOCK_COUNT + 1).
//
// FreeStack on Teensy3.2 is about 58K so 12 buffers which use ~6K is OK, could also use 28 buffer (~14K) as well if needed
// for 2 samples per 1mS each buffer takes 256mS to fill so 12 buffers give ~3sec of storage
// so can Use total of 12 512 byte buffers.
//const uint8_t BUFFER_BLOCK_COUNT = 12;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 128;  // Must be a power of two! and > (BUFFER_BLOCK_COUNT+1) so this good for upto 126 buffers

//==============================================================================
// test with 3 blocks at 27uS sampling
//0:43:30: adc conversions /sec: 37037 Max Buffers stored in one call to storeSampleBuffers() : 14
// Max SD latency (includes file close/open) : 47406uS Max file close/open latency : 17828uS
// adc0: 924 adc1: 931
// Missed Timers total: 0 Missed Samples in last block: 0 Total Missed Samples so far: 0
// >>>>>>> logt02.bin Total Blocks to write this file : 256000   Blocks written so far : 249150
// Total Blocks written : 761150

//==============================================================================

SdBaseFile binFile;  // current binary file


// from Teensy32AnalogLogger.h const size_t NAME_DIM = 13; // 8 + 3 + . + null
char binName[NAME_DIM] = FILE_BASE_NAME FILE_BASE_EXTRA;
char binName00[NAME_DIM] = FILE_BASE_NAME FILE_BASE_EXTRA;
char metaName[NAME_DIM] = FILE_BASE_NAME FILE_BASE_META_EXTRA ; // global storage for meta file name

bool processAllFiles = false; // set to true is e or c to process all the files
bool errorBlink = false; // set to true on some error
millisDelay blinkDelay;
uint32_t lastMissedSamples = 0;

block_t* emptyQueue[QUEUE_DIM];
uint8_t emptyHead;
uint8_t emptyTail;

block_t* fullQueue[QUEUE_DIM];
volatile uint8_t fullHead;  // volatile insures non-interrupt code sees changes.
uint8_t fullTail;

// Allocate buffer space. do it globally as we have the RAM available
block_t bufferBlocks[BUFFER_BLOCK_COUNT];
metadata_t _metadataBuffer;  // initialized by initAdc in setup();
metadata_t* metadataBufferPtr = &_metadataBuffer;

// queueNext assumes QUEUE_DIM is a power of two
inline uint8_t queueNext(uint8_t ht) {
  return (ht + 1) & (QUEUE_DIM - 1);
}
//==============================================================================
// Interrupt Service Routines

// Pointer to current buffer.
block_t* isrBuf;

// Need new buffer if true.
bool isrBufNeeded = true;

// missedSamples count, i.e. missed samples (i.e. 1 count == missed both ADC results for this sample time)
volatile uint32_t missedSamples = 0;
volatile uint32_t missedTimers = 0;
volatile uint32_t missedSamplesTotal = 0;

// Insure no timer events are missed.
volatile bool timerError = false;
volatile bool timerFlag = false;
volatile bool noBuffersError = false;

volatile bool samplingFlag = false; // set to true to start sampling false to stop
// timer runs all the time


void checkFileNameSizes() {
  bool nameSizeFail = false;
  //sizeof(FILE_BASE_NAME) etc include null so need to subtract 1
  if ((sizeof(FILE_BASE_NAME) - 1) > 6) {
    error("FILE_BASE_NAME must be <= 6 chars long.");
    nameSizeFail = true;
  }
  if ((sizeof(FILE_BASE_EXTRA) - 1) > 6) {
    error("FILE_BASE_EXTRA must be <= 6 chars long, including the '.' ");
    nameSizeFail = true;
  }
  if ((sizeof(FILE_BASE_META_EXTRA) - 1) > 4) {
    error("FILE_BASE_META_EXTRA must be <= 4 chars long, including the '.' ");
    nameSizeFail = true;
  }
  if (NAME_DIM < 13) {
    error("DIM_NAME must be >= 13 ");
    nameSizeFail = true;
  }
  if ( nameSizeFail ) {
    // stop here
    while (1) {}
  }
}

// This function will be called with the desired frequency
// start the measurement
void timer0_callback(void) {
  if (!samplingFlag) {
    return;
  }
  //  if (MONITOR_PIN >= 0) {
  //    digitalWriteFast(MONITOR_PIN, HIGH); //!digitalReadFast(MONITOR_PIN));
  //  }
  if (timerFlag) {
    timerError = true;
    if (missedTimers < 0XFFFFFFFF) {
      missedTimers++; // reset each block
    }
    // a missed timer means a missed sample as well
    if (missedSamples < 0XFFFFFFFF) {
      missedSamples++;
    }
    return;
  }
  timerFlag = true;
  adc->startSingleRead(PIN_LIST[0], ADC_0); // start ADC_0 first so it should complete before ADC_1 does
  // if we start ADC_1 before starting ADC_0, then the ADC_1 interrupt is called 4 times before ADC_0 completes, i.e. 3 extra times
  adc->startSingleRead(PIN_LIST[1], ADC_1); // start ADC_1 AFTER ADC_0
  // without sleep_cpu() SD latency ~200uS and ADC conversion time ~22uS
  //sleep_cpu(); // sleep here to improve ADC accuracy but may  increases max SD latency and  ADC conversion increases
}


//------------------------------------------------------------------------------

// loop here repeatedly calling this interrupt
// until adc0 is Complete and then saveSample which clears this interrupt
void adc1_isr() {
  if (adc->adc0->isComplete()) {
    // both complete  save values
    adc->adc1->readSingle(); // don't call this interrupt again
    saveSample(); // actually saveSample calls readSingle() anyway which clears this interrupt
  }
}

volatile  int adc0_reading;
volatile  int adc1_reading;

volatile uint32_t adcCount = 0;
void saveSample() {
  // Read ADC data.
  int adc0_data = adc->adc0->readSingle();
  int adc1_data = adc->adc1->readSingle();
  if (!samplingFlag) {
    // prevent and further updates to buffers
    return;
  }
  adc0_reading = adc0_data;
  adc1_reading = adc1_data;
  if (adcCount < 0XFFFFFFFF) {
    adcCount++; // count the saves
  }

  if (isrBufNeeded && emptyHead == emptyTail) {
    // no buffers - count missedSamples
    // missed this sample
    if (missedSamples < 0XFFFFFFFF) {
      missedSamples++;
    }
    // Avoid missed timer error.
    timerFlag = false;
    return;
  }

  // Check for buffer needed.
  if (isrBufNeeded) {
    // Remove buffer from empty queue.
    isrBuf = emptyQueue[emptyTail];
    emptyTail = queueNext(emptyTail);
    isrBuf->count = 0;
    isrBuf->missedSamples = 0;
    isrBufNeeded = false;
  }
  // Store ADC data.
  isrBuf->data[isrBuf->count++] = adc0_reading;
  isrBuf->data[isrBuf->count++] = adc1_reading;

  // Check for buffer full.
  if (isrBuf->count >= PIN_COUNT * SAMPLES_PER_BLOCK) {
    // Put buffer isrIn full queue.
    isrBuf->missedSamples = missedSamples; // save number of samples missed filling this buffer
    uint8_t tmp = fullHead;  // Avoid extra fetch of volatile fullHead.
    fullQueue[tmp] = isrBuf;
    fullHead = queueNext(tmp);

    // Set buffer needed and clear missedSampless.
    isrBufNeeded = true;
    missedSamples = 0;
  }
  timerFlag = false;
}

uint32_t noBlocksWritten = 0; // tracks blocs written to current file for truncate
uint32_t totalBlocksWritten = 0;
uint32_t totalSamplesWritten = 0;
uint32_t noBlocks = FILE_BLOCK_COUNT;
uint32_t maxFileOpenTime = 0;
int maxBuffersUsed = 0;
//------------------------------------------------------------------------------
// store on SD card buffer per call
// return true if buffer stored or nothing to do, else false if all files used OR if error
// Takes about 200uS to store data and ~15mS to move to next file
bool storeSampleBuffers() {
  uint32_t usecStart = micros();
  uint32_t fileOpenTime = 0;
  int fullBuffersCounter = 0;
  while (fullHead != fullTail) { // store all available buffers ~ 2.4mS for 12
    fullBuffersCounter++;
    if (!binFile.isOpen()) {
      return false;
    }
    // have a full block to save
    // else  (fullHead != fullTail)
    // Get address of block to write.
    block_t* pBlock = fullQueue[fullTail];

    // Write block to SD.
    if (!sd.card()->writeData((uint8_t*)pBlock)) {
      error("write data failed");
      //continue; // try next file
    }
    totalSamplesWritten += pBlock->count;

    // Add missedSamples to total and possibly light LED.
    if (pBlock->missedSamples > 0) {
      if (ERROR_LED_PIN >= 0) {
        digitalWrite(ERROR_LED_PIN, HIGH);
      }
      if ((0xFFFFFFFF - pBlock->missedSamples) > missedSamplesTotal) {
        missedSamplesTotal += pBlock->missedSamples;
      } else {
        missedSamplesTotal = 0XFFFFFFFF;
      }
    }

    // Move block to empty queue.
    emptyQueue[emptyHead] = pBlock;
    emptyHead = queueNext(emptyHead);
    fullTail = queueNext(fullTail);
    noBlocksWritten++;
    totalBlocksWritten++;
    if (noBlocksWritten >= noBlocks) {
      uint32_t fileOpenStart = micros();
      // this takes ~15mS to close old file and open new one
      closeBinFileForBlockWrite();
      if ( (fileCounter = setNextBinName(fileCounter)) == -1) {
        // all files have been used just stop here
        adcStop();
        return false; // need to truncate
      }  // else
      openBinFileForBlockWrite(); // open next file for next call to this method
      fileOpenTime = micros() - fileOpenStart;
      if (fileOpenTime > maxFileOpenTime) {
        maxFileOpenTime = fileOpenTime;
      }
    }
  }
  uint32_t usec = micros() - usecStart;
  if (usec > maxLatency) {
    maxLatency = usec;
  }
  if (fullBuffersCounter > maxBuffersUsed) {
    maxBuffersUsed = fullBuffersCounter;
  }
  return true;
}


//------------------------------------------------------------------------------
// sd_error_code_t in SdInfo.h
void fatalBlink() {
  errorBlink = true;
  if (
    (SD_CARD_ERROR_READ == sd.cardErrorCode()) ||
    (SD_CARD_ERROR_WRITE == sd.cardErrorCode()) ||
    (SD_CARD_ERROR_READ_TIMEOUT == sd.cardErrorCode()) ||
    (SD_CARD_ERROR_WRITE_TIMEOUT == sd.cardErrorCode())
  ) {
    // not fatal
  } else {
    fatalError = true;
  }
}
//==============================================================================

void initializeQueues() {
  // Initialize queues. -------------------------------
  emptyHead = emptyTail = 0;
  fullHead = fullTail = 0;

  // Put rest of buffers in the empty queue.
  for (uint8_t i = 0; i < BUFFER_BLOCK_COUNT; i++) {
    emptyQueue[emptyHead] = &bufferBlocks[i];
    emptyHead = queueNext(emptyHead);
  }
  //  end initialize queues --------------------------
}

//------------------------------------------------------------------------------
// initialize ADC and timer1
void adcInit(metadata_t* meta) {
  // reference can be ADC_REFERENCE::REF_3V3, ADC_REFERENCE::REF_1V2 (not for Teensy LC) or ADC_REFERENCE::REF_EXT.
  adc->setReference(ADC_REFERENCE::REF_3V3, ADC_0); // change all 3.3 to 1.2 if you change the reference to 1V2

  adc->setAveraging(ADC_AVGS); // set number of averages
  adc->setResolution(ADC_BITS); // set bits of resolution

  // it can be any of the ADC_CONVERSION_SPEED enum: VERY_LOW_SPEED, LOW_SPEED, MED_SPEED, HIGH_SPEED_16BITS, HIGH_SPEED or VERY_HIGH_SPEED
  // see the documentation for more information
  // additionally the conversion speed can also be ADACK_2_4, ADACK_4_0, ADACK_5_2 and ADACK_6_2,
  // where the numbers are the frequency of the ADC clock in MHz and are independent on the bus speed.
  adc->setConversionSpeed(ADC_CONVERSION_SPEED::MED_SPEED); // change the conversion speed
  // it can be any of the ADC_MED_SPEED enum: VERY_LOW_SPEED, LOW_SPEED, MED_SPEED, HIGH_SPEED or VERY_HIGH_SPEED
  adc->setSamplingSpeed(ADC_SAMPLING_SPEED::MED_SPEED); // change the sampling speed

  // adc->enableInterrupts(ADC_0);  // do not enable this interrupt, the ADC_1 interrupt is repeatedly called until both conversions complete
  adc->enableInterrupts(ADC_1);

  initializeQueues();

  meta->noBufferBlocks = BUFFER_BLOCK_COUNT;
  meta->pinCount = PIN_COUNT;
  for (int i = 0; i < PIN_COUNT; i++) {
    uint8_t pin = PIN_LIST[i];
    if (pin >= NUM_ANALOG_INPUTS) {
      error("Invalid Analog pin number");
      while (1) {
        ; // stop here
      }
    }
    meta->pinNumber[i] = pin;
  }

  meta->byteOrder = versionByteOrder; //0x0000ffff; // for version 0   Note NOT Unicode marker which is 0000fffe
  //  // Sample interval in uS
  meta->sampleInterval = SAMPLE_INTERVAL;
  meta->samplesPerBlock = SAMPLES_PER_BLOCK; // i.e. save PIN_COUNT * SAMPLES_PER_BLOCK values in a block == block count
  meta->adcBits = ADC_BITS;
  meta->adcAvgs = ADC_AVGS;
  meta->duration = 0;
  uint64_t fillFileTime_uS = (((uint64_t)meta->sampleInterval) * meta->samplesPerBlock * FILE_BLOCK_COUNT);
  uint64_t fillFileTime_mS = fillFileTime_uS / 1000;
  if (fillFileTime_mS > 0xffffffff) {
    fillFileTime_mS = 0xffffffff;
  }
  //serialIO->print(F(" Time to fill a file : "));   mStoHrMinSec((uint32_t)fillFileTime_mS);
  uint64_t allFilesFillTime_uS = fillFileTime_uS * FILE_COUNT;
  uint64_t allFilesFillTime_mS = allFilesFillTime_uS / 1000;
  if (allFilesFillTime_mS > 0xffffffff) {
    allFilesFillTime_mS = 0xffffffff;
  }
  // serialIO->print(F(" Time to fill ALL files : "));   mStoHrMinSec((uint32_t)allFilesFillTime_mS); serialIO->println();
  if (requestedRunTime > ((uint32_t)allFilesFillTime_mS)) {
    RUN_TIME_MS = (uint32_t)allFilesFillTime_mS;
    //serialIO->print(F(" Requested run time : "));   mStoHrMinSec(requestedRunTime);  serialIO->print(F(" exceeds time to fill all files. Run will be truncated.")); serialIO->println();
  }
  meta->runTime = RUN_TIME_MS;
  meta->maxBuffersUsed = 0;
  meta->maxLatency = 0;
  meta->maxFileOpenTime = 0;
  meta->missedTimers = 0;
  meta->missedSamplesTotal = 0;
  meta->totalBlocksWritten = 0;
  meta->totalSamplesWritten = 0;
  samplingFlag = false; // prevent ADC sampling when timer starts
  // timer0.end();
  loggingRunning = false;
}


//------------------------------------------------------------------------------
// enable ADC and timer1 interrupts
void adcStart() {
  if (haveAlreadyLoggedData) {
    return;
  }
  haveAlreadyLoggedData = true;
  //timer0.end(); // stop any old timer
  initializeQueues();
  maxBuffersUsed = 0;
  maxLatency = 0;
  maxFileOpenTime = 0;
  missedTimers = 0;
  missedSamplesTotal = 0;
  totalBlocksWritten = 0;
  totalSamplesWritten = 0;

  // initialize ISR
  isrBufNeeded = true;
  missedSamples = 0;
  timerError = false;
  timerFlag = false;
  noBuffersError = false;

  loggingRunning = true;
  mSstart = millis(); // start time
  samplingFlag = true; // release timer
}
//------------------------------------------------------------------------------
void adcStop() {
  samplingFlag = false;
  // serialIO->println(F(" Stopped ADC sampling."));
}

//----------------------------------------------------------------------------

void mStoHrMinSec(unsigned long _mS) {
  unsigned long secs = (unsigned long)(_mS / 1000);
  unsigned long mS = _mS - secs * 1000;
  unsigned long mins = secs / 60;
  secs = secs - mins * 60;
  unsigned long hrs = mins / 60;
  mins = mins - hrs * 60;
  serialIO->print(' ');
  serialIO->print(hrs); serialIO->print(':');
  if (mins <= 9) {
    serialIO->print('0');
  }
  serialIO->print(mins); serialIO->print(':');
  if (secs <= 9) {
    serialIO->print('0');
  }
  serialIO->print(secs); serialIO->print('.');
  if (mS <= 9) {
    serialIO->print(F("00"));
  } else if (mS <= 99) {
    serialIO->print('0');
  }
  serialIO->print(mS);
}


// -----------------------------------------------
// returns false when have reached FILE_COUNT files
// setNextBinName(0); initialize bin name
// returns -1 on all used

int setNextBinName(int count) {
  if (count < 0) {
    return -1;
  }
  if (count == 0) { // first time
    // initialize binName
    strcpy(binName, FILE_BASE_NAME);
    strcat(binName, FILE_BASE_EXTRA);
    return 1;
  }
  if (count > FILE_COUNT) {
    return -1; // finished
  }
  count++;
  // else set next name
  if (binName[BASE_NAME_SIZE + 1] != '9') {
    binName[BASE_NAME_SIZE + 1]++;
  } else {
    binName[BASE_NAME_SIZE + 1] = '0';
    if (binName[BASE_NAME_SIZE] == '9') {
      error("Can't create file name");
      return -1; // stop here
    }
    binName[BASE_NAME_SIZE]++;
  }
  return count;
}

// truncates partial files
void closeBinFileForBlockWrite() {
  if (!binFile.isOpen()) {
    return;
  }
  //  serialIO->println(F(" Closing block write sampling bin file."));

  if (!sd.card()->writeStop()) {
    serialIO->print(binName);
    error("writeStop failed");
    // keep going
  }
  if (noBlocksWritten != noBlocks) {
    // serialIO->print(F("Truncating file : ")); serialIO->println(binName);
    if (!binFile.truncate(512L * noBlocksWritten)) {
      error("Can't truncate file");
      // keep going to close
    }
  }
  binFile.close();
}

// -------------------------------------------------
// open current binName for block write
// creates/erases empty file if missing
// but create take time
void openBinFileForBlockWrite() {
  uint32_t bgnBlock, endBlock;

  if (!sd.exists(binName)) {
    createEmptyFile();
  }
  binFile.close();
  if (!binFile.open(binName, O_RDWR)) {
    serialIO->println(binName);
    error(" open failed");
    fatalBlink();
    return;
  }
  binFile.rewind();
  // Get the address of the file on the SD.
  if (!binFile.contiguousRange(&bgnBlock, &endBlock)) {
    error("contiguousRange failed");
    fatalBlink();
    binFile.close();
    return;
    //continue; // try next file
  }
  //*********************NOTE**************************************
  // NO SdFile calls are allowed while cache is used for raw writes
  //***************************************************************

  noBlocks = endBlock - bgnBlock + 1;
  // Start a multiple block write.
  if (!sd.card()->writeStart(bgnBlock, noBlocks)) {
    error("writeBegin failed");
    fatalBlink();
    binFile.close();
    return;
    // continue; // try next file
  }
  strcpy(metadataBufferPtr->currentFile, binName);
  noBlocksWritten = 0; // errors above here will abort saving to next file because block written at max
  //noBlocks = 1500; // !! TESTING !! remove for normal use
  // serialIO->print(F("!!! TESTING Setting noBlocks to : ")); serialIO->println(noBlocks);
}

//------------------------------------------------------------------------
bool createEmptyFile() {
  uint32_t bgnBlock, endBlock;
  serialIO->print(F("Creating new file : "));   serialIO->print(binName);
  // Remove file from current directory.
  if (sd.exists(binName)) {
    if (!sd.remove(binName)) {
      serialIO->println(binName);
      error(" remove failed");
      return false;
    }
  }
  //  unsigned long createStart = millis();
  binFile.close();
  if (!binFile.createContiguous(binName, 512 * FILE_BLOCK_COUNT)) {
    error("createContiguous failed. File Exists ? ");
    return false; // error OK here
  }
  // Get the address of the file on the SD.
  if (!binFile.contiguousRange(&bgnBlock, &endBlock)) {
    error("contiguousRange failed");
    binFile.close();
    return false;
    // error OK here
  }
  //*********************NOTE**************************************
  // NO SdFile calls are allowed while cache is used for raw writes
  //***************************************************************
  // Flash erase all data in the file.
  //serialIO->println(F("Erasing all data"));
  uint32_t bgnErase = bgnBlock;
  uint32_t endErase;
  while (bgnErase < endBlock) {
    endErase = bgnErase + ERASE_SIZE;
    if (endErase > endBlock) {
      endErase = endBlock;
    }
    if (!sd.card()->erase(bgnErase, endErase)) {
      error("erase failed");
      binFile.close();
      return false;
    }
    bgnErase = endErase + 1;
  }
  //serialIO->print(F("  CreateFile and erase time : ")); serialIO->print((millis() - createStart));  serialIO->println(F("mS"));
  binFile.close();
  return true;
}

//-----------------------------------------------------------------------------
uint32_t initializeElapsedTimeStart = 0;
// uses PROTECT_DATA to prevent overwrite

bool metadataExists = false;
bool metadataExistsSet = false;

bool metadataFileExists() {
  if (loggingRunning) {
    return true;
  } else if (!metadataExistsSet) {
    // serialIO->println("metadataFileExists");
    FatFile file;
    bool rtn = file.open(metaName, O_RDONLY);
    file.close();
    // serialIO->print(" ------------ end metadataFileExists returns : "); serialIO->println(rtn);
    metadataExistsSet = true;
    metadataExists = rtn;
    return rtn;
    //return sd.exists(metaName); // just calls open O_RDONLY to check
  } else {
    return metadataExists;
  }
}

bool createEmptyMetaFile() {
  bool rtn = true;
  FatFile metaFile;
  metadataExistsSet = false; // force check next time
  if (!metaFile.open(metaName, O_RDWR | O_CREAT)) {
    serialIO->println(" Error creating metaFile");
    removeMetadataFile();
    rtn = false;
  }
  metaFile.close();
  return rtn;
}

void removeMetadataFile() {
  if (!sd.remove(metaName)) {
    serialIO->println(" Error removing metaFile");
  } else {
    serialIO->println(" metadata File removed");
  }
  metadataExists = false;
  metadataExistsSet = false;
}

// if meta data file exists or any file already exist then cannot initialize
bool initializeFiles() {
  serialIO->print(" Initializing "); serialIO->print(FILE_COUNT);  serialIO->println(" files");
  //sd.vol()->cacheClear(); // does sync
  FatFile file;
  sd.chdir(true);
  if (PROTECT_DATA) { // should not get here
    if (metadataFileExists()) {
      serialIO->println(F("Data has already been logged."));
      fatalBlink();
      //fatalError = true;
      return false;
    }
  } else {
    removeMetadataFile();
  }
  if (PROTECT_DATA) { // should not get here
    if (sd.exists(binName00)) {
      serialIO->println(F("Already initialized."));
      return false;
    }
  }
  initializeElapsedTimeStart = millis();
  int i = 0; // number of files created so far
  while ((i = setNextBinName(i)) != -1) { // returns -1 after FILE_COUNT files
    // Create new file.
    if (PROTECT_DATA) {
      if (sd.exists(binName)) {
        serialIO->print(binName);
        error(" Log file Already exists.");
        fatalBlink();
        //fatalError = true;
        return false;
      }
    }
    if (!createEmptyFile()) {
      //sd.vol()->cacheClear(); // does sync
      return false; //uses binName deletes file if it exits
    }
    serialIO->print(F(" Elapsed time : ")); serialIO->print((millis() - initializeElapsedTimeStart)); serialIO->println(F("mS"));
    char c = checkInputForCmd(); // also sends output
    if (c == 's') {
      //  break; // ignore stop here
    }
    serialIO->flush(); // block here until current buffered output has been sent
  }
  // reset binName to start
  setNextBinName(0);
  //sd.vol()->cacheClear(); // does sync
  metadataExistsSet = false;
  return true;
}


// ----------------------------------------------------------------------------
void printMenu() {
  serialIO->println();
  if (loggingRunning) {
    serialIO->println("LOGGING DATA ..... ");
  }
  if (metadataFileExists() && (!loggingRunning)) {
    serialIO->println("DATA LOGGING COMPLETED! ");
  }
  if (fatalError) {
    serialIO->println(F(" FATAL Error."));
  }
  serialIO->print(F("Ver:")); serialIO->print(version); serialIO->print(F(SUB_VERSION)); serialIO->println(F("  enter one of the following commands:"));
  serialIO->println(F("? - current status and metadata"));
  if (loggingRunning) {
    serialIO->println(F("s - stop data logging"));
  } else {
    if (haveAlreadyLoggedData) {
      serialIO->println(F(" ** r - record ADC data ** not available. Data already exists"));
      serialIO->println(F(" ** i - initialize files ** not available. Data already exists"));
    } else {
      if (!fatalError) {
        if (metadataFileExists()) {
          serialIO->println(F(" ** r - record ADC data ** not available. Data already exists"));
          if (PROTECT_DATA) {
            serialIO->println(F(" ** i - initialize files ** not available. Data already exists"));
          } else {
            serialIO->println(F("i - initialize files")); // can re-run
          }
        } else { // no meta data
          if (sd.exists(binName00)) {
            serialIO->println(F("r - record ADC data"));
            if (!PROTECT_DATA) {
              serialIO->println(F("i - initialize files")); // can re-run
            }
          } else { // no 00.bin file
            serialIO->println(F("i - initialize files"));
          }
        }
      }
    }
    serialIO->println(F("l - list files"));
  }
  if (metadataFileExists()) {
    serialIO->println();
    if (loggingRunning) {
      serialIO->println(F(" LOGGING DATA...  Check with ? command"));
    } else {
      serialIO->println(F(" DATA HAS ALREADY BEEN LOGGED, check with ?"));
    }
  }
  serialIO->print('>');
  nonBlocking.clearInput();
}



//---------------------------------------------------
bool errorPinStateLOW = true;

void checkErrorBlink() {
  if (blinkDelay.justFinished()) {
    blinkDelay.restart();
    errorPinStateLOW = !errorPinStateLOW;
    if (errorBlink) {
      if (ERROR_LED_PIN > 0) {
        if (errorPinStateLOW) {
          digitalWrite(ERROR_LED_PIN, LOW);
        } else {
          digitalWrite(ERROR_LED_PIN, HIGH);
        }
      }
    }
  }
}

//--------------------------------------------------------
/**
  Stores last buffers and truncate files
  clear fullHead == fullTail
*/
void storeLastBuffersAndTruncate() {
  // serialIO->println(F("storeLastBuffersAndTruncate"));
  // save the current buffer
  uint8_t tmp = fullHead;  // Avoid extra fetch of volatile fullHead.
  fullQueue[tmp] = isrBuf;
  fullHead = queueNext(tmp);
  // serialIO->print(" Saving last samples "); serialIO->println(isrBuf->count);
  storeSampleBuffers(); // store the last ones
  closeBinFileForBlockWrite(); // truncates
  fullHead = fullTail = 0; // storeSampleBuffers will now always return true
  updateMetaData(metadataBufferPtr, true); // final update
  // did not get to the end of the files
  //setNextBinName return -1 on end of files
  while ((fileCounter = setNextBinName(fileCounter)) != -1) {
    if (sd.exists(binName)) {
      serialIO->print(" Removing unused file : "); serialIO->println(binName);
      serialIO->flush();
      if (!sd.remove(binName)) {
        serialIO->println(binName);
        error(" remove failed");
        // continue return;
      }
    }
  }
}

//-------------------------------------------------------

union {
  uint32_t num;
  uint8_t bytes[4];
} byteOrderUnion;

// -------------------------------------------------------
void outputMetaDataToSerial(metadata_t* meta) {
  float sampleRate = 1000000.0 / meta->sampleInterval;
  serialIO->println();
  //  serialIO->print(meta->duration); serialIO->println(F("mS"));
  if ((loggingRunning) || (haveAlreadyLoggedData)) {
    mStoHrMinSec(meta->duration);
  } else {
    mStoHrMinSec(0);
  }
  serialIO->print(" of "); mStoHrMinSec(meta->runTime);
  serialIO->println();
  serialIO->print(F(" Sample pins:"));
  for (uint8_t i = 0; i < meta->pinCount; i++) {
    serialIO->print(' ');
    serialIO->print(meta->pinNumber[i], DEC);
  }
  serialIO->print(F(" Byte order : "));
  byteOrderUnion.num = meta->byteOrder;
  if (byteOrderUnion.bytes[3] == 0xff) {
    serialIO->print(F("Big-Endian"));
  } else {
    serialIO->print(F("Little-Endian"));
  }
  serialIO->println();
  serialIO->print(F(" ADC bits: "));   serialIO->print(meta->adcBits);  serialIO->print(F(" ADC sample averages over : "));   serialIO->println(meta->adcAvgs);
  serialIO->print(F(" Sample Rate: "));   serialIO->print(sampleRate, 2); serialIO->print(F(" Sample interval: "));   serialIO->print(meta->sampleInterval); serialIO->println(F("uS"));
  serialIO->print(F(" Samples per block: "));   serialIO->print(meta->samplesPerBlock); serialIO->print(F(" Time to fill block : "));   serialIO->print(meta->sampleInterval * meta->samplesPerBlock); serialIO->println(F("uS"));
  uint64_t fillFileTime_uS = (((uint64_t)meta->sampleInterval) * meta->samplesPerBlock * FILE_BLOCK_COUNT);
  uint64_t fillFileTime_mS = fillFileTime_uS / 1000;
  if (fillFileTime_mS > 0xffffffff) {
    fillFileTime_mS = 0xffffffff;
  }
  serialIO->print(F(" Time to fill a file : "));   mStoHrMinSec((uint32_t)fillFileTime_mS);
  uint64_t allFilesFillTime_uS = fillFileTime_uS * FILE_COUNT;
  uint64_t allFilesFillTime_mS = allFilesFillTime_uS / 1000;
  if (allFilesFillTime_mS > 0xffffffff) {
    allFilesFillTime_mS = 0xffffffff;
  }
  serialIO->print(F(" Time to fill ALL files : "));   mStoHrMinSec((uint32_t)allFilesFillTime_mS); serialIO->println();
  if (requestedRunTime > ((uint32_t)allFilesFillTime_mS)) {
    serialIO->print(F(" Requested run time : "));   mStoHrMinSec(requestedRunTime);  serialIO->print(F(" exceeds time to fill all files. Run will be truncated.")); serialIO->println();
  }

  serialIO->print(F(" Max SD latency (includes file close/open) : ")); serialIO->print(meta->maxLatency);  serialIO->print(F("uS")); serialIO->print(F(" Max file close/open latency : ")); serialIO->print(meta->maxFileOpenTime);  serialIO->println(F("uS"));
  serialIO->print(F(" Number of buffer blocks: "));   serialIO->print(meta->noBufferBlocks);
  serialIO->print(F(" Time to fill ALL block buffers : ")); serialIO->print(meta->noBufferBlocks * meta->sampleInterval * meta->samplesPerBlock); serialIO->println(F("uS"));
  serialIO->print(F(" Max number of buffers saved in call to storeSampleBuffers() : "));   serialIO->println(meta->maxBuffersUsed);
  serialIO->print(F(" Missed Timers total : ")); serialIO->print(meta->missedTimers);  serialIO->print(F(" Total Missed Samples so far: ")); serialIO->println(meta->missedSamplesTotal);
  serialIO->print(F(" Total Blocks written : ")); serialIO->print(meta->totalBlocksWritten);
  serialIO->print(F(" Total Samples written : ")); serialIO->print(meta->totalSamplesWritten / meta->pinCount);
  serialIO->print(F(" covering : ")); mStoHrMinSec( (((unsigned long)(meta->totalSamplesWritten / meta->pinCount)) * meta->sampleInterval) / 1000); serialIO->println();
  serialIO->print(F(" Current File : ")); serialIO->println(meta->currentFile);

}

// ---------------------------------------------------------
/**
  Assumes input has been already been initialized by adcInit()
  update metafile with results from last run, or with most recent results
  uses values in global mSstart for runtime, maxBuffersUsed, maxLatency, maxFileOpenTime, missedTimers, totalMissedSamples, totalBlocksWritten
  updates   metadata_t metadataBuffer;
*/
void updateMetaData(metadata_t* meta, bool updateDuration) {
  // closeBinFileForBlockWrite updates last file closed
  if (updateDuration) {
    meta->duration = millis() - mSstart;
  }
  meta->maxBuffersUsed = maxBuffersUsed;
  meta->maxLatency = maxLatency;
  meta->maxFileOpenTime = maxFileOpenTime;
  meta->missedTimers = missedTimers;
  meta->missedSamplesTotal = missedSamplesTotal;
  meta->totalBlocksWritten = totalBlocksWritten;
  meta->totalSamplesWritten = totalSamplesWritten;
}

// -----------------------------------
bool updateMetaDataFile(metadata_t* meta) {
  //updateMetaData(meta,updateDuration);
  return writeMetadataFile(meta);
}

// -------------------------------------------
/*
  Open metadata file .met and read in data
  updates arg
*/
void readMetaDataFile(metadata_t* metaBufferPtr) {
  // serialIO->println("readMetaDataFile");
  if (!metadataFileExists()) {
    error("Log Data has NOT been run yet.");
    return;
  }
  SdBaseFile metaFile;  // current binary file
  //metaFile.close();
  // create meta file

  if (!metaFile.open(metaName, O_RDONLY)) {
    serialIO->println(metaName);
    error(" open failed");
    return;
  }
  if (metaFile.read((uint8_t*)metaBufferPtr , sizeof(metadata_t)) != sizeof(metadata_t)) {
    error("Read metadata failed");
  }
  metaFile.close();
}

//-------------------------------------
/**
  write current  metadataBuffer
*/
bool writeMetadataFile(metadata_t* metaBufferPtr) {
  // serialIO->println("writeMetadataFile");
  SdBaseFile metaFile;  // current binary file
  bool rtn = true;
  // serialIO->println(metaName);
  if (!metadataFileExists()) {
    serialIO->println(metaName);
    error(" does not exist ");
    return false;
  }
  // create meta file if missing else update
  if (!metaFile.open(metaName, O_RDWR)) {
    serialIO->println(metaName);
    error(" open failed");
    return false;
  }
  //  metaFile.rewind();
  int bytesWritten =  metaFile.write((uint8_t*)metaBufferPtr, sizeof(metadata_t));
  if (sizeof(metadata_t) != bytesWritten) {
    serialIO->println(metaName);    serialIO->print(F(" byte written:")); serialIO->println(bytesWritten);
    error(" Error writing meta data.");
    rtn = false;
  }
  metaFile.close();
  //sd.vol()->cacheClear(); // does sync
  return rtn;
}
// ----------------------------------------------------------------------

void processingOnADCstop() {
  storeLastBuffersAndTruncate(); // clears buffers so storeSampleBuffers() returns true from now on
  updateMetaDataFile(metadataBufferPtr); // ignore error here
  loggingRunning = false;
}

//------------------------------------------------------------------------------
void setup(void) {

  version = (uint16_t)(versionByteOrder >> 16);
  if (ERROR_LED_PIN >= 0) {
    pinMode(ERROR_LED_PIN, OUTPUT);
    digitalWrite(ERROR_LED_PIN, LOW); // gnd
  }
  if (ERROR_LED_GND_PIN >= 0) {
    pinMode(ERROR_LED_GND_PIN, OUTPUT);
    digitalWrite(ERROR_LED_GND_PIN, LOW); // gnd
  }
  if (MONITOR_PIN >= 0) {
    pinMode(MONITOR_PIN, OUTPUT);
    digitalWrite(MONITOR_PIN, LOW); // off
  }
  COM_SERIAL.begin(baudRate); // connected to bluetooth module
  //#define COM_SERIAL at the top of this file sets which COM port is used for cmds
#ifdef DEBUG
  Serial.begin(115200);
#endif
  for (int i = 10; i > 0; i--) {
#ifdef DEBUG
    Serial.println(i);
#endif
    delay(500);
  }

  blinkDelay.start(200);

#ifdef DEBUG
  Serial.print(F("FreeStack: "));
  Serial.println(FreeStack());
#endif
  // Initialize at the highest speed supported by the board that is
  // not over 50 MHz. Try a lower speed if SPI errors occur.
  if (!sd.begin(SD_CS_PIN, SD_SCK_MHZ(50))) {
    sd.initErrorPrint();
    fatalBlink();
    //fatalError = true;
    return;
  }

  checkFileNameSizes();

  if (PIN_COUNT > 2) {
    error("Too many Sample pins (2 max)");
    fatalBlink();
    //fatalError = true;
    return;
  }

  for (int i = 0; i < PIN_COUNT; i++) {
    pinMode(PIN_LIST[i], INPUT);
  }
  adc->adc0->recalibrate();
  adc->adc1->recalibrate();
  delay(1000);
#ifdef DEBUG
  Serial.println();
  Serial.println(F("ls:"));
  sd.ls(&Serial, LS_DATE | LS_SIZE);
  Serial.println();
#endif

  //#define COM_SERIAL at the top of this file sets which COM port is used for cmds
  serialIO = bufferedStream.connect(&COM_SERIAL); // either Serial for USB or Serial1 for bluetooth module
  nonBlocking.connect(serialIO);
  lineBuf[0] = '\0';
  adcInit(metadataBufferPtr); // initialize buffers and fill in meta data  check runtime against files
  printMenu();

  // start the timer, if it's not possible, the  return value  will be false
  if (!timer0.begin(timer0_callback, SAMPLE_INTERVAL) ) {
    error("Could not start timer0)");
    fatalBlink();
    //fatalError = true;
    return;
  }

}

void processRunCmd() {
  if (loggingRunning) {
    serialIO->println(" Already logging data.");
    return;
  }

  if (haveAlreadyLoggedData || metadataFileExists() ) {
    serialIO->print(F(" Already have logged data : ")); serialIO->print(metaName); serialIO->println(F(" exists."));
    return;
  }

  adcInit(metadataBufferPtr); // initialize buffers and fill in meta data  check runtime against files
  // create metadata file if missing
  if (!createEmptyMetaFile()) {
    serialIO->println(" Error creating metadata file");
    return;
  }

  if (!writeMetadataFile(metadataBufferPtr)) {
    serialIO->println(" Error writing metadata file");
  }
  //serialIO->println(F("Wrote metadata"));
  fileCounter = setNextBinName(0); // reset filename to first file
  openBinFileForBlockWrite();
  //serialIO->println(F("Opened BinFile for block write"));
  outputMetaDataToSerial(metadataBufferPtr);
  adcStart();
  printTimer.start(PRINT_INTERVAL);
  stopTimer.start(RUN_TIME_MS); // stop sampling after this time
}

// -----------------------------------------------------
void handleInput(char c) {
  // handle input
  if (c == 0) {
    return;
  }
  // else
  if (loggingRunning &&
      ((c != '?') && (c != 's'))
     ) {
    c = 0; // invalid cmd
  }

  if (c == '?') {
    if ((!loggingRunning) && metadataFileExists()) {
      metadata_t buf;
      readMetaDataFile(&buf);
      outputMetaDataToSerial(&buf);
    } else {
      updateMetaData(metadataBufferPtr, loggingRunning);
      outputMetaDataToSerial(metadataBufferPtr);
    }

  } else if (c == 's') {
    if (loggingRunning)  {
      adcStop();
      serialIO->println(F(" Stopping Logging and removing unused files."));
      serialIO->flush();
    } else { //else ignore
      serialIO->println(F(" Not logging "));
    }

  } else if (c == 'l') {
    serialIO->println(F("ls:"));
    sd.ls(serialIO, LS_DATE | LS_SIZE);

  } else if (c == 'i') {
    if (haveAlreadyLoggedData) {
      serialIO->println(" Have already logged data.");
    } else if (!initializeFiles()) {
      serialIO->println(" Error initializing files");
    }

  } else if (c == 'r') {
    if (!sd.exists(binName00)) {
      serialIO->println(" Need to run initialize first.");
      return;
    } else {

      if (haveAlreadyLoggedData) {
        serialIO->println(" Have already logged data.");
      } else {
        processRunCmd();
      }
    }

  } else {
    serialIO->println(F("Invalid cmd"));
  }
  if (fatalError) {
    serialIO->println(F(" !! Fatal Error !!"));
  }
  printMenu();
}

// --------------------------------------------------------

char checkInputForCmd() {
  char c = 0;
  int len = nonBlocking.readInputLine(lineBuf, lineBufLen, true);
  if (len > 0) {
    c = lineBuf[0];
    lineBuf[0] = '\0';
    serialIO->println(); // terminate output line using our idea of end of line
  }
  return c;
}
//------------------------------------------------------------------------------
void loop(void) {
  checkErrorBlink();

  if (loggingRunning && ((!samplingFlag) || (!storeSampleBuffers()) )) {
    // store all full buffer each call. Returns true if buffer stored, else false if all files used OR if error
    adcStop(); // incase of file error or full files
    stopTimer.stop();
    processingOnADCstop();
    metadata_t buf;
    readMetaDataFile(&buf); // read it back in
    outputMetaDataToSerial(&buf);
    serialIO->println();
    serialIO->println(F("ls:"));
    sd.ls(&bufferedStream, LS_DATE | LS_SIZE);
    serialIO->println();
    printMenu();
  }


  if (printTimer.justFinished()) {
    printTimer.repeat();
    if (samplingFlag) { //only print this if still sampling
      serialIO->println();
      updateMetaData(metadataBufferPtr, true);
      serialIO->print(" Elapsed Run time : "); mStoHrMinSec(metadataBufferPtr->duration);
      serialIO->print(" of "); mStoHrMinSec(metadataBufferPtr->runTime);
      if (RUN_TIME_MS < requestedRunTime) {
        serialIO->print(" limited by number of files.");
      }
      serialIO->println();
      if (metadataBufferPtr->missedSamplesTotal != 0) {
        serialIO->print(F(" Error!! missed samples so far : ")); serialIO->println(metadataBufferPtr->missedSamplesTotal);
      }
      if ((timerError) || (noBuffersError)) {
        if (timerError) {
          timerError = false;
          serialIO->print(F(" Error!! missed timer"));
        }
        if (noBuffersError) {
          noBuffersError = false;
          serialIO->print(F(" Error!! out of buffers"));
        }
        serialIO->println();
      }
    }
  }
  // TESTING only
  if (stopTimer.justFinished()) {
    adcStop();
  }
  char c = checkInputForCmd();
  handleInput(c);
}
#else  // not teensy 3.2
#error This program is only for Teensy 3.2
#endif  // __MK20DX256__
